#include <torch/extension.h>
#include <ATen/ATen.h>
#include <ATen/cuda/CUDAContext.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <math.h>
#include <vector>

#define CHECK_CUDA(x) TORCH_CHECK(x.is_cuda(), #x " must be a CUDA tensor")
#define CHECK_CONTIGUOUS(x) TORCH_CHECK(x.is_contiguous(), #x " must be contiguous")
#define CHECK_INPUT(x) CHECK_CUDA(x); CHECK_CONTIGUOUS(x)

// This fused kernel combines vectorized memory accesses (float4 loads/stores) with grid-stride loops
// to handle arbitrary group sizes. It fuses Swish activation, Group Normalization (with warp-level reduction),
// and HardSwish activation in two phases: first, a reduction phase to compute mean/variance using vectorized loads
// where possible; second, a normalization and activation phase that re-computes Swish and applies the affine transform.

__global__ void fused_vec_gridstride_kernel(
    const float* __restrict__ input,
    float* __restrict__ output,
    const float* __restrict__ gamma,
    const float* __restrict__ beta,
    int N, int C, int D, int H, int W,
    int groups, float eps
) {
    // Each block processes one group from one sample
    int n = blockIdx.x;
    int g = blockIdx.y;

    int channels_per_group = C / groups;
    int group_elements = channels_per_group * D * H * W;
    int base = n * (C * D * H * W) + g * group_elements;

    const int VECTOR_SIZE = 4;
    int aligned_elements = (group_elements / VECTOR_SIZE) * VECTOR_SIZE;  // number of elements that can be processed vectorized

    int tid = threadIdx.x;
    int blockSize = blockDim.x;

    // Phase 1: Compute partial sums and sums-of-squares for the Swish activated values
    float local_sum = 0.0f;
    float local_sumsq = 0.0f;

    // Use vectorized loads for the aligned portion
    for (int i = tid * VECTOR_SIZE; i < aligned_elements; i += blockSize * VECTOR_SIZE) {
        int idx = base + i;
        // Load four consecutive floats assuming alignment
        float4 vec = *reinterpret_cast<const float4*>(&input[idx]);
        float x0 = vec.x, x1 = vec.y, x2 = vec.z, x3 = vec.w;
        float sw0 = x0 / (1.f + expf(-x0));
        float sw1 = x1 / (1.f + expf(-x1));
        float sw2 = x2 / (1.f + expf(-x2));
        float sw3 = x3 / (1.f + expf(-x3));
        local_sum += sw0 + sw1 + sw2 + sw3;
        local_sumsq += sw0 * sw0 + sw1 * sw1 + sw2 * sw2 + sw3 * sw3;
    }

    // Handle any tail elements that are not vectorized
    for (int i = aligned_elements + tid; i < group_elements; i += blockSize) {
        int idx = base + i;
        float x = __ldg(&input[idx]);
        float sw = x / (1.f + expf(-x));
        local_sum += sw;
        local_sumsq += sw * sw;
    }

    // Warp-level reduction using shuffles
    unsigned int mask = 0xffffffff;
    for (int offset = warpSize/2; offset > 0; offset /= 2) {
        local_sum += __shfl_down_sync(mask, local_sum, offset);
        local_sumsq += __shfl_down_sync(mask, local_sumsq, offset);
    }

    // Shared memory for storing warp-level partial results
    extern __shared__ float shared[]; // shared memory: first half for sums, second half for sumsq
    int numWarps = (blockSize + warpSize - 1) / warpSize;
    float* warpSums = shared;
    float* warpSumsSq = shared + numWarps;

    int warpId = tid / warpSize;
    int laneId = tid % warpSize;
    if(laneId == 0) {
        warpSums[warpId] = local_sum;
        warpSumsSq[warpId] = local_sumsq;
    }
    __syncthreads();

    float total_sum = 0.f;
    float total_sumsq = 0.f;
    if(tid < numWarps) {
        total_sum = warpSums[tid];
        total_sumsq = warpSumsSq[tid];
    }
    __syncthreads();
    if(tid == 0) {
        for (int i = 1; i < numWarps; i++) {
            total_sum += warpSums[i];
            total_sumsq += warpSumsSq[i];
        }
        warpSums[0] = total_sum;  // store final reduction in shared memory
        warpSumsSq[0] = total_sumsq;
    }
    __syncthreads();

    total_sum = warpSums[0];
    total_sumsq = warpSumsSq[0];
    float mean = total_sum / group_elements;
    float var = total_sumsq / group_elements - mean * mean;
    float inv_std = rsqrtf(var + eps);

    // Phase 2: Recompute the activation and apply normalization + HardSwish
    // Use vectorized stores for the aligned part
    for (int i = tid * VECTOR_SIZE; i < aligned_elements; i += blockSize * VECTOR_SIZE) {
        int idx = base + i;
        float4 in_vec = *reinterpret_cast<const float4*>(&input[idx]);
        float4 out_vec;
        #pragma unroll
        for (int j = 0; j < VECTOR_SIZE; j++) {
            float x = ((&in_vec.x)[j]);
            float sw = x / (1.f + expf(-x));
            int local_channel = (i + j) / (D * H * W);
            int global_channel = g * channels_per_group + local_channel;
            float norm = (sw - mean) * inv_std;
            float y = norm * __ldg(&gamma[global_channel]) + __ldg(&beta[global_channel]);
            float hs = y * fminf(fmaxf(y + 3.f, 0.f), 6.f) / 6.f;
            (&out_vec.x)[j] = hs;
        }
        *reinterpret_cast<float4*>(&output[idx]) = out_vec;
    }

    // Process remaining tail elements
    for (int i = aligned_elements + tid; i < group_elements; i += blockSize) {
        int idx = base + i;
        float x = __ldg(&input[idx]);
        float sw = x / (1.f + expf(-x));
        int local_channel = i / (D * H * W);
        int global_channel = g * channels_per_group + local_channel;
        float norm = (sw - mean) * inv_std;
        float y = norm * __ldg(&gamma[global_channel]) + __ldg(&beta[global_channel]);
        output[idx] = y * fminf(fmaxf(y + 3.f, 0.f), 6.f) / 6.f;
    }
}

// The forward function first applies a conv_transpose3d then launches the fused kernel
torch::Tensor forward(
    torch::Tensor x,
    int stride,
    int padding,
    int groups,
    float eps,
    torch::Tensor conv_transpose,
    torch::Tensor conv_transpose_bias,
    torch::Tensor group_norm_weight,
    torch::Tensor group_norm_bias
) {
    CHECK_INPUT(x);
    CHECK_INPUT(conv_transpose);
    CHECK_INPUT(conv_transpose_bias);
    CHECK_INPUT(group_norm_weight);
    CHECK_INPUT(group_norm_bias);

    // Perform the transposed convolution
    x = torch::conv_transpose3d(x, conv_transpose, conv_transpose_bias, stride, padding);
    torch::Tensor output = torch::empty_like(x);

    int N = x.size(0);
    int C = x.size(1);
    int D = x.size(2);
    int H = x.size(3);
    int W = x.size(4);

    // Each block processes one (sample, group) pair
    dim3 grid(N, groups);
    int threads = 256;
    int numWarps = (threads + 31) / 32;
    size_t shared_mem = 2 * numWarps * sizeof(float);

    fused_vec_gridstride_kernel<<<grid, threads, shared_mem>>>(
        x.data_ptr<float>(),
        output.data_ptr<float>(),
        group_norm_weight.data_ptr<float>(),
        group_norm_bias.data_ptr<float>(),
        N, C, D, H, W,
        groups, eps
    );

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Fused vectorized and grid-stride kernel for Swish, GroupNorm, and HardSwish");
}
